/**********************************************************************
*  2011 Microchip Technology Inc.
*
* FileName:        main.c
* Dependencies:    Header (.h) files if applicable, see below
* Processor:       PIC32MX795F512L
* Compiler:        MPLAB C32 v1.11a or higher
*
* SOFTWARE LICENSE AGREEMENT:
* Microchip Technology Incorporated ("Microchip") retains all ownership and 
* intellectual property rights in the code accompanying this message and in all 
* derivatives hereto.  You may use this code, and any derivatives created by 
* any person or entity by or on your behalf, exclusively with Microchip's
* proprietary products.  Your acceptance and/or use of this code constitutes 
* agreement to the terms and conditions of this notice.
*
* CODE ACCOMPANYING THIS MESSAGE IS SUPPLIED BY MICROCHIP "AS IS".  NO 
* WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED 
* TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A 
* PARTICULAR PURPOSE APPLY TO THIS CODE, ITS INTERACTION WITH MICROCHIP'S 
* PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN ANY APPLICATION. 
*
* YOU ACKNOWLEDGE AND AGREE THAT, IN NO EVENT, SHALL MICROCHIP BE LIABLE, WHETHER 
* IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF STATUTORY DUTY), 
* STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, 
* PUNITIVE, EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF 
* ANY KIND WHATSOEVER RELATED TO THE CODE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN 
* ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.  TO THE FULLEST EXTENT 
* ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO 
* THIS CODE, SHALL NOT EXCEED THE PRICE YOU PAID DIRECTLY TO MICROCHIP SPECIFICALLY TO 
* HAVE THIS CODE DEVELOPED.
*
* You agree that you are solely responsible for testing the code and 
* determining its suitability.  Microchip has no obligation to modify, test, 
* certify, or support the code.
************************************************************************/

#include "..\h\WM8960Drv.h"
#include "..\h\tone.h"
#include "..\h\DemoDefines.h"
#include "..\h\GFXFunctions.h"

/*************************************************************************
 * TODO: Refer to Step 1 in the Lab manual
 *************************************************************************/

        //<- Add Code here
        //<- Add Code here

/************************************************************************
 * The following section configures the PIC32 clock and oscillator
 * settings. The Audio Development board has a 8MHz crystal on board.
 * The device is configured to operate at 80MHz. This example does 
 * not use USB and so the USP PLL is swithced off.
 ************************************************************************/

#pragma config UPLLEN   = OFF,      FPLLMUL  = MUL_20,  UPLLIDIV = DIV_2
#pragma config FPLLIDIV = DIV_2,    FPLLODIV = DIV_1,   FPBDIV   = DIV_1
#pragma config FWDTEN   = OFF,      WDTPS    = PS1,     OSCIOFNC = OFF
#pragma config POSCMOD  = XT,       IESO     = OFF,     FSOSCEN  = OFF
#pragma config FNOSC    = PRIPLL,   CP       = OFF ,    BWP      = OFF
#pragma config PWP      = OFF,      ICESEL   = ICS_PGx2   


/*********************************************************
 * TODO: Refer to Step 2 in the Lab Manual
 *********************************************************/

#define FRAME_SIZE      //<-Add Code here
                        //<-Add Code here

/*********************************************************
 * The following array contains samples for a 400Hz 
 * tone. The tone samples are in fractional format (Q15)
 * and were generated using dsPICWorks tool. The actual
 * array is contained in the tone.h file (contained
 * in the h folder of the project).
 **********************************************************/

extern int tone[40];

/********************************************************
 * The switchDebounce variable tracks when the user
 * switches should be polled for activity. It is modified
 * by the Debounce Timer ISR.
 ********************************************************/

volatile BOOL switchDebounce = FALSE;

/*****************************************************
 * TODO: Refer to Step 3 in the Lab Manual
 *****************************************************/

#define FILTER_ORDER          //<- Add Code here
#define FILTER_INPUT_SIZE     //<- Add Code here

        //<-Add Code here
        //<-Add Code here
        //<-Add Code here
        //<-Add Code here
        //<-Add Code here

int main(void)
{
    INT i, toneSampleCount;
    WM8960_CODEC * pCodecHandle;
    BOOL filterEnable = TRUE;

    /************************************************
     * Do some system and application initialization
     * here.
     ************************************************/

    SYSTEMConfigPerformance(SYSTEM_CLOCK);
    INTEnableSystemMultiVectoredInt();
    INIT_FILTER_ONOFF_SWITCH();
    INIT_DEBOUNCE_TIMER();
    GFXDisplayInit();
    GFXUpdateSwitch(filterEnable);

    /*************************************************
     * Get a handle to the Codec Driver. Note that 
     * this requires dynamic memory to allocated
     * The project options has this setup already.
     *************************************************/
    pCodecHandle = WM8960CodecOpen(O_RDWR);

    if(pCodecHandle == NULL)
    {
        /* Need more memory to run program */
        while(1);
    }

    /******************************************************
     * Configure the codec. Enable the microphone. Setup
     * the sampling rate for 16KHz. Set up the headphone
     * volume
     ******************************************************/

    WM8960CodecConfigSampleRate(pCodecHandle, SAMPLE_RATE_16000_HZ  );
    WM8960CodecEnableMicrophone(pCodecHandle, TRUE);
    WM8960CodecSetMicGain(pCodecHandle, 0x3F);
    WM8960CodecSetMicBoost(pCodecHandle, 0x1);
    WM8960CodecSetLeftVolume(pCodecHandle,0x78);
    WM8960CodecSetRightVolume(pCodecHandle,0x78);
    WM8960CodecStartAudio(pCodecHandle,TRUE);

    toneSampleCount = 0;

    /****************************************************
     * TODO: Refer to step 4 in the lab manual
     ***************************************************/
            
            //<- Add Code here

    /***************************************************
     * TODO: Refer to step 5 in the Lab Manual
     ***************************************************/

            //<- Add code here

    while(1)
    {
        WM8960CodecRead(pCodecHandle,audio,FRAME_SIZE);

        /**********************************************
         * Add the tone to input speech signal. The 
         * microphone signal is contained in the left
         * channel of the input signal.
         **********************************************/

        for(i = 0; i < FRAME_SIZE; i ++)
        {
            filterInput[i] = audio[i].leftChannel + tone[toneSampleCount++];
            toneSampleCount = (toneSampleCount >= 40) ? 0 : toneSampleCount;
        }

        /***************************************************
         * TODO: Refer to step 6 in the Lab Manual
         ***************************************************/

                //<- Add code here

        if(filterEnable == TRUE)
        {

            /************************************************* 
             * If the switch is enabled, then send the filter
             * output to the codec
             *************************************************/

            for(i = 0; i < FRAME_SIZE; i++)
            {
                audio[i].rightChannel = filterOutput[i];
                audio[i].leftChannel = filterOutput[i];
            }
        }
        else
        {
            /*************************************************
             * If the switch is disabled then send the filter
             * input to the codec. This bypasses the filter
             *************************************************/

            for(i = 0; i < FRAME_SIZE; i++)
            {
                audio[i].rightChannel = filterInput[i];
                audio[i].leftChannel = filterInput[i];
            }

        }    

        WM8960CodecWrite(pCodecHandle,audio,FRAME_SIZE);

        if(switchDebounce == TRUE)
        {
            /****************************************
             * Time to check the swithces
             ****************************************/

            switchDebounce = FALSE;
            if(CheckFilterSwitch() == TRUE)
            {
                /******************************************
                 * If the switch is activated, then toggle
                 * the filter activation.
                 ******************************************/

                filterEnable = (filterEnable == TRUE) ? FALSE : TRUE;
            }
            GFXUpdateSwitch(filterEnable);

        }    
    }

    return 0;
}

BOOL CheckFilterSwitch(void)
{
    /**********************************************
     * This function will check the filter enable
     * switch. It should be called every debounce
     * period. The function will consider a switch
     * press valid only if the switch was released
     * before being pressed again
     **********************************************/
     
    static INT prevValue = 1;
    
    INT currentSwitchValue = FILTER_ONOFF;

    if(currentSwitchValue == prevValue) 
    {
        return(FALSE);
    }
    
    prevValue = currentSwitchValue;
    
    if(currentSwitchValue == 0)
    {
        return(TRUE);
    }
    
    return(FALSE);
}        

void __attribute__((vector(4), interrupt(ipl3), nomips16)) TimerISR(void)
{
    /****************************************
     * Clear the interrupt flag and indicate
     * that the switch debounce period has 
     * expired.
     *****************************************/

    IFS0CLR = 0x10;
    switchDebounce = TRUE;

}

