#ifndef _RTLL_H_
#define _RTLL_H_


/**************************************
 * This data structure defines exported
 * library functions in the module 
 * header.
 *************************************/

typedef struct
{
    const char* procString;
    const void* proc;
}T_PROC_DCPT;   // procedure descriptor    



/************************************
 * This data structure defines the
 * module header. Each library should
 * have one such structure associated 
 * with it.
 ************************************/
typedef struct
{
	const char 				*module_name;				// name of the module
    void 				   (*libInit)(void*);    	  	// Library start-up code
    const unsigned int      nProcs;               		// Number of procedures
    const T_PROC_DCPT*      procTbl;              		// pointer to the procedures   
}T_MODULE_DYN_HDR;

/************************************
 * This is the address of the module
 * header in program flash memory.
 ************************************/

#define HELIX_LIB_MODULE_HEADER_ADDRESS 0x9D06E000

void * dlopen(char *libString, void* libLoadPoint, void * libOpenParam);
void * dlsym(void* libHandle, char* funcName);

#endif

