/**********************************************************************
*  2011 Microchip Technology Inc.
*
* FileName:        AudioUSBTasks.c
* Dependencies:    Header (.h) files if applicable, see below
* Processor:       PIC32MX795F512L
* Compiler:        MPLAB C32 v1.11a or higher
*
* SOFTWARE LICENSE AGREEMENT:
* Microchip Technology Incorporated ("Microchip") retains all ownership and 
* intellectual property rights in the code accompanying this message and in all 
* derivatives hereto.  You may use this code, and any derivatives created by 
* any person or entity by or on your behalf, exclusively with Microchip's
* proprietary products.  Your acceptance and/or use of this code constitutes 
* agreement to the terms and conditions of this notice.
*
* CODE ACCOMPANYING THIS MESSAGE IS SUPPLIED BY MICROCHIP "AS IS".  NO 
* WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED 
* TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A 
* PARTICULAR PURPOSE APPLY TO THIS CODE, ITS INTERACTION WITH MICROCHIP'S 
* PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN ANY APPLICATION. 
*
* YOU ACKNOWLEDGE AND AGREE THAT, IN NO EVENT, SHALL MICROCHIP BE LIABLE, WHETHER 
* IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF STATUTORY DUTY), 
* STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, 
* PUNITIVE, EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF 
* ANY KIND WHATSOEVER RELATED TO THE CODE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN 
* ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.  TO THE FULLEST EXTENT 
* ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO 
* THIS CODE, SHALL NOT EXCEED THE PRICE YOU PAID DIRECTLY TO MICROCHIP SPECIFICALLY TO 
* HAVE THIS CODE DEVELOPED.
*
* You agree that you are solely responsible for testing the code and 
* determining its suitability.  Microchip has no obligation to modify, test, 
* certify, or support the code.
************************************************************************/
#include "RTLL.h"

/**************************
 * USB and MSD Host Includes
 **************************/
#include "USB/usb.h"
#include "USB/usb_host_msd.h"
#include "USB/usb_host_msd_scsi.h"
#include "MDD File System/FSIO.h"
#include "GFXFunctions.h"
/******************************
 * WM8960 Codec Driver 
 ******************************/
#include "..\h\WM8960Drv.h"

/*******************************
 * File IO 
 *******************************/
volatile BOOL deviceAttached;
FSFILE * mp3File;
int searchResult;
SearchRec rec;
INT nRead = 0;

/********************************
* WM8960 Audio Codec 
**********************************/
WM8960_CODEC * pCodecHandle;
STEREO_AUDIO_DATA txBuffer[WM8960DRV_TX_BUFFER_SIZE_STEREO_WORD];
INT nOutputSamples = 0;


/********************************************
 * TODO: Refer to Step 1 of the Lab Manual
 ********************************************/
            //<- Add Code Here	
            //<- Add Code Here
            //<- Add Code Here
            //<- Add Code Here
	


/*****************************************
 * TODO: Refer to Step 2 of the Lab Manual
 *****************************************/
            //<-Add Code here

/************************************
* MP3 Decoder status and variables.
*************************************/

INT bytesLeft, nRead, err;
INT offset, outOfData, eofReached;

/********************************
 * This stucture maps to the 
 * frame information structure
 * used by Helix MP3 decoder
 *******************************/
 
 typedef struct {
	int bitrate;
	int nChans;
	int samprate;
	int bitsPerSample;
	int outputSamps;
	int layer;
	int version;
} FRAME_INFO;

FRAME_INFO frameInfo;


/*****************************************
 * TODO: Refer to Step 3 of the Lab Manual
 *****************************************/

                //<-Add Code here

/*****************************************
 * TODO: Refer to Step 4 of the Lab Manual
 *****************************************/
 
                
void * (*MP3InitDecoder)(void);	 //<- Review this
void * (*MP3FreeDecoder)(void * hMP3Decoder);
int    (*MP3Decode)(void * hMP3Decoder, 
        unsigned char **inbuf, int *bytesLeft, short *outbuf, int useSize);
void   (*MP3GetLastFrameInfo)(void * hMP3Decoder, FRAME_INFO *mp3FrameInfo);
int    (*MP3GetNextFrameInfo)(void * hMP3Decoder, FRAME_INFO *mp3FrameInfo, unsigned char *buf);
int    (*MP3FindSyncWord)(unsigned char *buf, int nBytes);
void   (*RegisterMalloc)(void * fp);
void   (*RegisterFree)(void * fp);

/**********************************
 * These are possible application 
 * states.
 *********************************/

typedef enum 
{
    IDLE_STATE,
    DEVICE_ATTACHED,
    SEARCH_MP3_FILES,
    OPEN_MP3_FILE,
    FILE_OPEN,
    FRAME_SYNC_FIND,
    FILE_READ,
    MP3_DECODE,
    PLAYBACK,
    FILE_CLOSE,
    PLAYBACK_PAUSE,
    PLAYBACK_STOP,
    SYSTEM_HALT
    
}APPLICATION_STATE;

APPLICATION_STATE applicationState,previousAppState;

static int FillReadBuffer(unsigned char *readBuf, unsigned char *readPtr, int bufSize, int bytesLeft, FSFILE *infile)
{
    /* This function will move bytesLeft bytes from
     * readPtr to the top of readBuf. bufSize is the
     * size of readBuf. It then reads bufSize - bytesLeft bytes
     * from infile and appends these bytes to readBuf. 
     * If readBuf is not full, then remaing bytes are
     * set to zero. The total number of bytes read from 
     * infile is returned.
     */

    int nRead;

    /* move last, small chunk from end of buffer to start, then fill with new data */
    memmove(readBuf, readPtr, bytesLeft);				
    nRead = FSfread( readBuf + bytesLeft, 1, bufSize - bytesLeft, infile);
    /* zero-pad to avoid finding false sync word after last frame (from old data in readBuf) */
    if (nRead < bufSize - bytesLeft)
        memset(readBuf + bytesLeft + nRead, 0, bufSize - bytesLeft - nRead);	

    return nRead;
}

/****************************************************************************
  Function:
    BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event,
                void *data, DWORD size )

  Summary:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.

  Description:
    This is the application event handler.  It is called when the stack has
    an event that needs to be handled by the application layer rather than
    by the client driver.  If the application is able to handle the event, it
    returns TRUE.  Otherwise, it returns FALSE.

  Precondition:
    None

  Parameters:
    BYTE address    - Address of device where event occurred
    USB_EVENT event - Identifies the event that occured
    void *data      - Pointer to event-specific data
    DWORD size      - Size of the event-specific data

  Return Values:
    TRUE    - The event was handled
    FALSE   - The event was not handled

  Remarks:
    The application may also implement an event handling routine if it
    requires knowledge of events.  To do so, it must implement a routine that
    matches this function signature and define the USB_HOST_APP_EVENT_HANDLER
    macro as the name of that function.
  ***************************************************************************/

BOOL USB_ApplicationEventHandler( BYTE address, USB_EVENT event, void *data, DWORD size )
{
    switch( event )
    {
        case EVENT_VBUS_REQUEST_POWER:
            // The data pointer points to a byte that represents the amount of power
            // requested in mA, divided by two.  If the device wants too much power,
            // we reject it.
            return TRUE;

        case EVENT_VBUS_RELEASE_POWER:
            // Turn off Vbus power.
            // The PIC24F with the Explorer 16 cannot turn off Vbus through software.

            // This means that the device was removed
            // Indicate device was detached and make the
            // file count 0. Change focus to first item
            // in the list. Close the mp3 file if open
            // Stop and clear the codec
            
            deviceAttached = FALSE;
            return TRUE;
            break;

        case EVENT_HUB_ATTACH:
            return TRUE;
            break;

        case EVENT_UNSUPPORTED_DEVICE:
            return TRUE;
            break;

        case EVENT_CANNOT_ENUMERATE:
            //UART2PrintString( "\r\n***** USB Error - cannot enumerate device *****\r\n" );
            return TRUE;
            break;

        case EVENT_CLIENT_INIT_ERROR:
            //UART2PrintString( "\r\n***** USB Error - client driver initialization error *****\r\n" );
            return TRUE;
            break;

        case EVENT_OUT_OF_MEMORY:
            //UART2PrintString( "\r\n***** USB Error - out of heap memory *****\r\n" );
            return TRUE;
            break;

        case EVENT_UNSPECIFIED_ERROR:   // This should never be generated.
            //UART2PrintString( "\r\n***** USB Error - unspecified *****\r\n" );
            return TRUE;
            break;

        default:
            break;
    }

    return FALSE;
}

void AudioUSBInit(void)
{
    deviceAttached = FALSE;
    AD1PCFGbits.PCFG5 = 1;

    //Initialize the USB stack
    USBInitialize(0);

    /************************************************
     * Initialize the WM8760 codec for playback only
     * and load up the start up values. The playback
     * does not start here.
     ************************************************/

    pCodecHandle = WM8960CodecOpen(O_WRONLY);
    WM8960CodecConfigSampleRate(pCodecHandle, SAMPLE_RATE_44100_HZ  );
    WM8960CodecEnableMicrophone(pCodecHandle, TRUE);
    WM8960CodecSetMicGain(pCodecHandle, 0x3F);
    WM8960CodecSetMicBoost(pCodecHandle, 0x1);
    WM8960CodecSetLeftVolume(pCodecHandle,0x78);
    WM8960CodecSetRightVolume(pCodecHandle,0x78);

    /* Initial application state */
    applicationState = IDLE_STATE;

    /* Get a handle to the MP3 decoder and
     * obtain an entry point to the decoder
     * Library functions. */

    /*******************************************
     * TODO: Refer to Step 5 of the Lab Manual
     *******************************************/

                    //<-Add Code here

    if( hMP3DecoderLibrary == 0)
    {
        /* There was problem while opening the
         * MP3 Decoder Library. For now we will
         * stop here. */

        while(1);
    }

    /*******************************************
     * TODO: Refer to Step 6 of the Lab Manual
     *******************************************/
    MP3InitDecoder      = dlsym(hMP3DecoderLibrary, "HelixMP3InitDecoder");
            //<- Add Code here
            //<- Add Code here
            //<- Add Code here
    MP3GetNextFrameInfo = dlsym(hMP3DecoderLibrary, "HelixMP3GetNextFrameInfo");
    MP3FindSyncWord     = dlsym(hMP3DecoderLibrary, "HelixMP3FindSyncWord");
    RegisterMalloc      = dlsym(hMP3DecoderLibrary, "HelixMP3RegMalloc");
    RegisterFree        = dlsym(hMP3DecoderLibrary, "HelixMP3RegFree");

    if(MP3InitDecoder      == 0) while(1);
    if(MP3FreeDecoder      == 0) while(1);
    if(MP3Decode           == 0) while(1);
    if(MP3GetLastFrameInfo == 0) while(1);
    if(MP3GetNextFrameInfo == 0) while(1);
    if(MP3FindSyncWord     == 0) while(1);
    if(RegisterMalloc      == 0) while(1);
    if(RegisterFree        == 0) while(1);

    RegisterMalloc(malloc);
    RegisterFree(free);

    /*********************************************
     * TODO: Refer to Step 7 of the Lab Manual
     *********************************************/

                     //<- Add Code here
    if(hMP3Decoder == 0)
    {
        /* The MP3 Decoder could not be initialized
         * Stop here. */

        while(1);
    }

}


/*******************************************************************
 * Function: AudioTasks()
 * Description: This function implements the application state machine
 ********************************************************************/

void AudioUSBTasks(void)
{

    //USB stack process function
    USBTasks();

    if(deviceAttached == FALSE )
    {
        /* Rather than have to check for 
         * device detach on every step of the
         * state machine, we check for device
         * at the very beginning. */


        applicationState = IDLE_STATE;
        WM8960CodecStartAudio(pCodecHandle,FALSE);

        if(mp3File != NULL)
        {
            FSfclose(mp3File);
            mp3File = NULL;
        }
    }

    switch(applicationState)
    {
        case(IDLE_STATE):
            {
                /* In this state the application is 
                 * waiting for USB Mass Storage device
                 * to be connected. This is the initial
                 * state of the application. The application
                 * also reaches here if the USB device 
                 * has been detached. */

                if(USBHostMSDSCSIMediaDetect() == TRUE)
                {
                    /* Device is Attached */
                    deviceAttached = TRUE;
                    applicationState = DEVICE_ATTACHED;
                }
                break;
            }

        case(DEVICE_ATTACHED):
            {
                /* Initialize the File system after
                 * device is attached. If this does 
                 * work for some reason, the application
                 * will keep trying unless the device
                 * is detached. */

                if((FSInit()) == TRUE)
                {
                    /* File system is initialized. Now search
                     * MP3 files. Indicate Device is attached. */
                    GFXStatusIndicate(DRIVE_FOUND);
                    applicationState = OPEN_MP3_FILE;
                }
                else
                {
                    /* The media not supported or corrupted
                    */
                    GFXStatusIndicate(DRIVE_ERROR);
                    applicationState = SYSTEM_HALT;
                }

                break;

            }

        case(OPEN_MP3_FILE):
            {
                /* The application reaches here after it
                 * has opened  atleast one mp3 file (irrespective
                 * of whether it was played out at all).
                 */

                mp3File =  FSfopen("test.mp3","r"); 
                if(mp3File != NULL)
                {
                    /* Found a file and its open */
                    applicationState = FILE_OPEN ;
                }
                else
                {
                    GFXStatusIndicate(FILE_NOT_FOUND);
                    applicationState = SYSTEM_HALT;
                }

                break;
            }

        case(FILE_OPEN):
            {

                /* Intialize the MP3 decoder variables
                 * as we will start decoding a new file.
                 * The details of the MP3 file are obtained
                 * and the read pointer is set up to point
                 * to the start of the MP3 frame.*/

                /* At this point, the MP3 decoder is already
                 * initialized. Reset all the state variables */ 

                bytesLeft = 0;
                outOfData = 0;
                eofReached = 0;
                readPtr = readBuf;
                nRead = 0;
                applicationState = FRAME_SYNC_FIND;
                break;
            }    

        case(FRAME_SYNC_FIND):
            {

                /* Read the mp3 file. Fill the first buffer of 
                 * data from the input file. */

                bytesLeft = FSfread(readBuf,1,READBUF_SIZE,mp3File);

                /* If end of file is reached then close this file */
                if(FSfeof(mp3File) != 0)
                {
                    applicationState = FILE_CLOSE; 
                    break;
                }


                /*****************************************************
                 * TODO: Refer to Step 8 of the Lab Manual
                 ****************************************************/

                                  //<-Add Code here
                
                if (offset < 0) 
                {
                    /* This means this frame does not contain
                     * the frame sync. Get another frame. */

                    readPtr = readBuf;
                    applicationState = FRAME_SYNC_FIND;
                    break;

                }

                /* Update the read pointer and number of bytes
                 * left in the current frame. */

                readPtr += offset;
                bytesLeft -= offset;

                /* Determine the audio format from the first frame header
                 * The information in mp3FrameInfo can be displayed on the
                 * display. */

                MP3GetNextFrameInfo(hMP3Decoder, &frameInfo, readPtr); 
                if(frameInfo.samprate != 44100 || frameInfo.nChans != 2)
                {
                    /* This means that the MP3 file is either a
                     * mono audio stream or not sampled at 44100Hz.
                     * We currently dont want to handle these type 
                     * of files even if the decoder can decode
                     * them */

                    applicationState = FILE_CLOSE;
                    break;

                }           

                /* Start the audio codec.*/ 
                WM8960CodecStartAudio(pCodecHandle,TRUE);

                /* Start reading the MP3 file.*/
                applicationState = FILE_READ;
                break;
            }

        case(FILE_READ):
            {
                /* Read some more data is the number of bytes
                 * left is less than a frame size and we have
                 * not reached the end of file.
                 */

                if (bytesLeft < READBUF_SIZE && !eofReached) 
                {
                    nRead = FillReadBuffer(readBuf, readPtr, 
                            READBUF_SIZE, bytesLeft, mp3File);
                    bytesLeft += nRead;
                    readPtr = readBuf;
                    if (nRead == 0)
                        eofReached = 1;
                }

                /* Decode the frame */
                applicationState = MP3_DECODE;


                break;
            }

        case(MP3_DECODE):
            {
                short * decoderOutput;
                
                /*****************************************************
                 * TODO: Refer to Step 9 of the Lab Manual
                 ****************************************************/
                
              
                                //<- Add Code here

                MP3GetLastFrameInfo(hMP3Decoder, &frameInfo);
                nOutputSamples = (frameInfo.outputSamps/2);
                applicationState = PLAYBACK;
                if (err) 
                {
                    /* error occurred */
                    switch (err) 
                    {
                        case -1:
                            applicationState = FILE_CLOSE;
                            break;
                        case -2:
                            /* More data is needed. Read the file. */
                            applicationState = FILE_READ;
                            break;
                        case -3:
                            applicationState = FILE_CLOSE;
                            break;
                        case -6:
                            applicationState = FRAME_SYNC_FIND;
                            break;
                        default:
                            applicationState = FILE_CLOSE;
                            break;
                    }
                } 

                break;
            }

        case(PLAYBACK):
            {
                /* In this state an stereo audio frame is
                 * available to play. Play back the frame
                 * and then read the file for more 
                 * data */

                STEREO_AUDIO_DATA * audio = (STEREO_AUDIO_DATA *)outBuf;
                WM8960CodecWrite(pCodecHandle, audio,nOutputSamples);
                applicationState = FILE_READ;
                break;


            }

        case (SYSTEM_HALT):
            break;

        case(FILE_CLOSE):
            {
                /* The application would have reached here if
                 * there is an error opening the file or the
                 * playback is done. In either case try the 
                 * next file in the list
                 */

                WM8960CodecStartAudio(pCodecHandle,FALSE);
                WM8960CodecClearTxBuffer(pCodecHandle);
                FSfclose(mp3File);
                mp3File = NULL;
                applicationState = OPEN_MP3_FILE;
                break;
            }
        default:
            while(1);
    }

}



