/**********************************************************************
*  2010 Microchip Technology Inc.
*
* FileName:        main.c
* Dependencies:    Header (.h) files if applicable, see below
* Processor:       PIC32MX
* Compiler:        MPLAB C32 
*
* SOFTWARE LICENSE AGREEMENT:
* Microchip Technology Incorporated ("Microchip") retains all ownership and 
* intellectual property rights in the code accompanying this message and in all 
* derivatives hereto.  You may use this code, and any derivatives created by 
* any person or entity by or on your behalf, exclusively with Microchip's
* proprietary products.  Your acceptance and/or use of this code constitutes 
* agreement to the terms and conditions of this notice.
*
* CODE ACCOMPANYING THIS MESSAGE IS SUPPLIED BY MICROCHIP "AS IS".  NO 
* WARRANTIES, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED 
* TO, IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A 
* PARTICULAR PURPOSE APPLY TO THIS CODE, ITS INTERACTION WITH MICROCHIP'S 
* PRODUCTS, COMBINATION WITH ANY OTHER PRODUCTS, OR USE IN ANY APPLICATION. 
*
* YOU ACKNOWLEDGE AND AGREE THAT, IN NO EVENT, SHALL MICROCHIP BE LIABLE, WHETHER 
* IN CONTRACT, WARRANTY, TORT (INCLUDING NEGLIGENCE OR BREACH OF STATUTORY DUTY), 
* STRICT LIABILITY, INDEMNITY, CONTRIBUTION, OR OTHERWISE, FOR ANY INDIRECT, SPECIAL, 
* PUNITIVE, EXEMPLARY, INCIDENTAL OR CONSEQUENTIAL LOSS, DAMAGE, FOR COST OR EXPENSE OF 
* ANY KIND WHATSOEVER RELATED TO THE CODE, HOWSOEVER CAUSED, EVEN IF MICROCHIP HAS BEEN 
* ADVISED OF THE POSSIBILITY OR THE DAMAGES ARE FORESEEABLE.  TO THE FULLEST EXTENT 
* ALLOWABLE BY LAW, MICROCHIP'S TOTAL LIABILITY ON ALL CLAIMS IN ANY WAY RELATED TO 
* THIS CODE, SHALL NOT EXCEED THE PRICE YOU PAID DIRECTLY TO MICROCHIP SPECIFICALLY TO 
* HAVE THIS CODE DEVELOPED.
*
* You agree that you are solely responsible for testing the code and 
* determining its suitability.  Microchip has no obligation to modify, test, 
* certify, or support the code.
************************************************************************/
#include "WM8960Drv.h"
#include "GFXFunctions.h"
#include "DemoDefines.h"

/******************************************************************
 * TODO: Refer to Step 1 in the Lab Manual
 ******************************************************************/

#include "speex/speex.h"   //<- Add Code here

/************************************************************************
 * The following section configures the PIC32 clock and oscillator
 * settings. The Audio Development board has a 8MHz crystal on board.
 * The device is configured to operate at 80MHz. This example does 
 * not use USB and so the USP PLL is swithced off.
 ************************************************************************/

#pragma config UPLLEN   = OFF,      FPLLMUL  = MUL_20,  UPLLIDIV = DIV_2
#pragma config FPLLIDIV = DIV_2,    FPLLODIV = DIV_1,   FPBDIV   = DIV_1
#pragma config FWDTEN   = OFF,      WDTPS    = PS1,     OSCIOFNC = OFF
#pragma config POSCMOD  = XT,       IESO     = OFF,     FSOSCEN  = OFF
#pragma config FNOSC    = PRIPLL,   CP       = OFF ,    BWP      = OFF
#pragma config PWP      = OFF,      ICESEL   = ICS_PGx2   

/********************************************************************
 * TODO: Refer to Step 2 in the Lab Manual
 ********************************************************************/

#define SPEEX_DECODER_PERCEPTUAL_ENHANCER 1     //<-Add Code Here
#define SPEEX_ENCODER_QUALITY 3                 //<-Add Code Here
#define MAX_ENCODED_PACKET_SIZE 100             //<-Add Code Here
#define MEMORY_BUFFER_SIZE 64000                //<-Add Code Here
#define FRAME_SIZE 160                          //<-Add Code Here

/*******************************************************************
 * The audio array will store the audio data to be read and written
 * to the codec driver. switchDebounce tracks when the switches
 * should be scanned. buffer is the byte memory where thecompressed
 * data is stored.
 *******************************************************************/

STEREO_AUDIO_DATA audio[FRAME_SIZE];
volatile BOOL switchDebounce = FALSE;
BYTE buffer[MEMORY_BUFFER_SIZE];

int main(void)
{
    UINT32      quality = SPEEX_ENCODER_QUALITY;
    UINT32      perceptualEnhancer = SPEEX_DECODER_PERCEPTUAL_ENHANCER;
    INT16       rawData[FRAME_SIZE];
    INT         nRead,nEncBytes;
    INT         i,currentLength;
    INT         recordBufferIndex   = 0;
    INT         playbackBufferIndex = 0;
    BOOL        record          = FALSE;
    BOOL        wasRecording    = FALSE;
    BOOL        playback        = FALSE;
    BOOL        wasPlayback     = FALSE;
    BOOL        fileIsOpen      = FALSE;

    WM8960_CODEC * pCodecHandle;

    /***********************************************************
     * TODO: Refer to Step 3 in the Lab Manual
     ***********************************************************/
    
    void        *encoder;       //<- Add code here
    void        *decoder;       //<- Add code here
    SpeexBits   encoderBits;    //<- Add code here
    SpeexBits   decoderBits;    //<- Add code here

    /************************************************
     * Do some system and application initialization
     * here.
     ************************************************/

    SYSTEMConfigPerformance(SYSTEM_CLOCK);
    INTEnableSystemMultiVectoredInt();

    GFXDisplayInit();
    INIT_RECORD_SWITCH();
    INIT_PLAYBACK_SWITCH();
    INIT_DEBOUNCE_TIMER();

    /*************************************************
     * Get a handle to the Codec Driver. Note that 
     * this requires dynamic memory to allocated
     * The project options has this setup already.
     *************************************************/

    pCodecHandle = WM8960CodecOpen(O_RDWR);

    if(pCodecHandle == NULL)
    {
        /* Need more memory to run program */
        while(1);
    }

    
    speex_bits_init(&encoderBits);                  

    /***********************************************************
     * TODO: Refer to Step 4 in the Lab Manual
     ***********************************************************/

    encoder = speex_encoder_init(&speex_nb_mode);   //<- Add Code here
    if(encoder == 0)
    {
        while(1);
    }
    speex_encoder_ctl(encoder, SPEEX_SET_QUALITY, &quality);    

    speex_bits_init(&decoderBits);                  

    /***********************************************************
     * TODO: Refer to Step 5 in the Lab Manual
     ***********************************************************/

    decoder = speex_decoder_init(&speex_nb_mode);   //<- Add code here
    if(decoder == 0)
    {
        while(1);
    }
    speex_decoder_ctl(decoder, SPEEX_SET_ENH, &perceptualEnhancer); 

    /******************************************************
     * Configure the codec. Enable the microphone. Setup
     * the sampling rate for 8KHz. Set up the headphone
     * volume
     ******************************************************/

    WM8960CodecConfigSampleRate(pCodecHandle, SAMPLE_RATE_8000_HZ  );
    WM8960CodecEnableMicrophone(pCodecHandle, TRUE);
    WM8960CodecSetMicGain(pCodecHandle, 0x3F);
    WM8960CodecSetMicBoost(pCodecHandle, 0x1);
    WM8960CodecSetLeftVolume(pCodecHandle,0x78);
    WM8960CodecSetRightVolume(pCodecHandle,0x78);
    WM8960CodecStartAudio(pCodecHandle,TRUE);

    while(1)
    {
        /* Read a frame of audio */
        WM8960CodecRead(pCodecHandle,audio,FRAME_SIZE);

        for(i = 0; i < FRAME_SIZE; i++)
        {
            rawData[i] = audio[i].leftChannel;
        }


        if(record == TRUE)
        {
            if(recordBufferIndex >= (MEMORY_BUFFER_SIZE - 10))
            {
                record = FALSE;
                GFXRecPlaybackIndicate(record,playback);
            }
            wasRecording = TRUE;


            speex_bits_reset(&encoderBits); 

            /******************************************************
             * TODO: Refer to Step 6 of the Lab Manual
             ******************************************************/

            speex_encode_int(encoder,rawData,&encoderBits); //<-Add code here
            
            nEncBytes = speex_bits_write(&encoderBits,
                        &buffer[recordBufferIndex],MAX_ENCODED_PACKET_SIZE);    
            recordBufferIndex += nEncBytes;
                
        }

        if(record == FALSE)
        {
            if(wasRecording == TRUE)
            {
                wasRecording = FALSE;
                currentLength = recordBufferIndex;
                recordBufferIndex = 0;
            }
        }


        if(playback == TRUE)
        {
            wasPlayback = TRUE;
            if(currentLength == 0)
            {
                playback = FALSE;
            }    
            
            
            speex_bits_read_from(&decoderBits, &buffer[playbackBufferIndex], nEncBytes); 

            /******************************************************
             * TODO: Refer to Step 7 of the Lab Manual
             ******************************************************/

            speex_decode_int(decoder, &decoderBits, rawData);   //<- Add code here

            playbackBufferIndex += nEncBytes;
            playbackBufferIndex = (playbackBufferIndex >= currentLength) ? 0 : playbackBufferIndex;
        }

        if(playback == FALSE)
        {
            if(wasPlayback == TRUE)
            {
                playbackBufferIndex = 0;
                wasPlayback = FALSE;
                GFXRecPlaybackIndicate(record,playback);
            }
        }

        for(i = 0; i < FRAME_SIZE; i++)
        {
            audio[i].leftChannel = rawData[i];
            audio[i].rightChannel = rawData[i];
        }     
        WM8960CodecWrite(pCodecHandle,audio,FRAME_SIZE);

        if(switchDebounce == TRUE)
        {
            /********************************************
             * Check the switches here. Pressing a switch
             * toggles operation. If playback is pressed
             * while record is active, the record will be
             * de-activated and playback will start. The
             * opposite holds true as well.
             ********************************************/

            switchDebounce = FALSE;

            if(CheckRecordSwitch() == TRUE)
            {
                record = (record == TRUE) ? FALSE : TRUE;
                playback = (record == TRUE) ? FALSE : playback;
                if(record == TRUE)
                {
                    recordBufferIndex = 0;
                    currentLength = 0;
                }    
                
            }

            if(CheckPlaybackSwitch() == TRUE)
            {
                playback = (playback == TRUE) ? FALSE : TRUE;
                record = (playback == TRUE) ? FALSE : record;
                if(playback == TRUE)
                {
                    playbackBufferIndex = 0;
                }    
            }

            /* Update the display */

            GFXRecPlaybackIndicate(record, playback);    

        }

    } 

    return 0;
}


void __attribute__((vector(4), interrupt(ipl3), nomips16)) TimerISR(void)
{
    /******************************************
     * Clear the interrupt flag and indicate
     * that the switch debounce period has 
     * expired. 
     ******************************************/

    IFS0CLR = 0x10;
    switchDebounce = TRUE;

}

BOOL CheckRecordSwitch(void)
{
    /**************************************************************
     * This function will check the record switch. It should be
     * called every debounce period. The function will consider
     * a switch press valid only if the switch was released before
     * being pressed again.
     **************************************************************/
     
    static INT prevValue = 1;
    
    INT currentSwitchValue = RECORD_SWITCH;

    if(currentSwitchValue == prevValue) 
    {
        return(FALSE);
    }
    
    prevValue = currentSwitchValue;
    
    if(currentSwitchValue == 0)
    {
        return(TRUE);
    }
    
    return(FALSE);
}        

BOOL CheckPlaybackSwitch()
{
    /**************************************************************
     * This function will check the playback switch. It should be
     * called every debounce period. The function will consider
     * a switch press valid only if the switch was released before
     * being pressed again.
     **************************************************************/
    
     
    static INT prevValue = 1;
    
    INT currentSwitchValue = PLAYBACK_SWITCH;

    if(currentSwitchValue == prevValue) 
    {
        return(FALSE);
    }
    
    prevValue = currentSwitchValue;
    
    if(currentSwitchValue == 0)
    {
        return(TRUE);
    }
    
    return(FALSE);
}        


