/*************************************************************************
 * Processor-specific object file.  Contains SFR definitions.
 *************************************************************************/
INPUT("processor.o")

/*************************************************************************
 * For interrupt vector handling
 *************************************************************************/
PROVIDE(_vector_spacing = 0x00000001);
_ebase_address = 0x9FC01000;

/*************************************************************************
 * Memory Address Equates
 * _RESET_ADDR      -- Reset Vector
 * _BEV_EXCPT_ADDR  -- Boot exception Vector
 * _DBG_EXCPT_ADDR  -- In-circuit Debugging Exception Vector
 * _DBG_CODE_ADDR   -- In-circuit Debug Executive address
 * _DBG_CODE_SIZE   -- In-circuit Debug Executive size
 * _GEN_EXCPT_ADDR  -- General Exception Vector
 *************************************************************************/
_RESET_ADDR              = 0xBFC00000;
_BEV_EXCPT_ADDR          = 0xBFC00380;
_DBG_EXCPT_ADDR          = 0xBFC00480;
_DBG_CODE_ADDR           = 0xBFC02000;
_DBG_CODE_SIZE           = 0xFF0     ;
_GEN_EXCPT_ADDR          = _ebase_address + 0x180;

/*************************************************************************
 * Memory Regions
 *
 * Memory regions without attributes cannot be used for orphaned sections.
 * Only sections specifically assigned to these regions can be allocated
 * into these regions.
 *************************************************************************/
MEMORY
{
    /* The Helix MP3 Library code starts from 0x9D06E100. This accomodates
     * the module header that is placed at 0x9D06E000 and is 256 bytes in
     * length. The Helix MP3 library code is allocated approximately 73K
     * of program flash memory. */
    kseg0_program_mem    (rx)  : ORIGIN = 0x9D06E100, LENGTH = 0x11F00
    kseg0_boot_mem             : ORIGIN = 0x9FC00490, LENGTH = 0x970
    exception_mem              : ORIGIN = 0x9FC01000, LENGTH = 0x1000
    kseg1_boot_mem             : ORIGIN = 0xBFC00000, LENGTH = 0x490
    debug_exec_mem             : ORIGIN = 0xBFC02000, LENGTH = 0xFF0
    config3                    : ORIGIN = 0xBFC02FF0, LENGTH = 0x4
    config2                    : ORIGIN = 0xBFC02FF4, LENGTH = 0x4
    config1                    : ORIGIN = 0xBFC02FF8, LENGTH = 0x4
    config0                    : ORIGIN = 0xBFC02FFC, LENGTH = 0x4
    
    /* The Helix MP3 Library code data area starts from 0xA001FF00 and is
     * 256 bytes long. This memory is enough to accomadate any of the 
     * global variables and data structure in the library. The library 
     * uses application memory space for heap and stack. XC32 linker
     * requests for additional stack memory. We will make this 2000 bytes
     */

    kseg1_data_mem       (w!x) : ORIGIN = 0xA001F830, LENGTH = 0x7D0
    sfrs                       : ORIGIN = 0xBF800000, LENGTH = 0x100000
    configsfrs                 : ORIGIN = 0xBFC02FF0, LENGTH = 0x10
}

/*************************************************************************
 * Configuration-word sections
 *************************************************************************/
SECTIONS
{
  .config_BFC02FF0 : {
    KEEP(*(.config_BFC02FF0))
  } > config3
  .config_BFC02FF4 : {
    KEEP(*(.config_BFC02FF4))
  } > config2
  .config_BFC02FF8 : {
    KEEP(*(.config_BFC02FF8))
  } > config1
  .config_BFC02FFC : {
    KEEP(*(.config_BFC02FFC))
  } > config0
}
