/*********************************************************************
 *
 *       Run Time Library Loading(RTLL)- System Loader Framework         
 *         
 *
 *********************************************************************
 * FileName:        system_loader.c
 * Dependencies:
 * Processor:       PIC32
 *
 * Complier:        MPLAB C32
 *                  MPLAB IDE
 * Company:         Microchip Technology, Inc.
 *
 * Software License Agreement
 * Microchip Audio Library  PIC32 Software.
 * Copyright  2008 Microchip Technology Inc.  All rights reserved.
 * 
 * Microchip licenses the Software for your use with Microchip microcontrollers
 * and Microchip digital signal controllers pursuant to the terms of the
 * Non-Exclusive Software License Agreement accompanying this Software.
 *
 * SOFTWARE AND DOCUMENTATION ARE PROVIDED AS IS WITHOUT WARRANTY
 * OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION,
 * ANY WARRANTY OF MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS
 * FOR A PARTICULAR PURPOSE.
 * MICROCHIP AND ITS LICENSORS ASSUME NO RESPONSIBILITY FOR THE ACCURACY,
 * RELIABILITY OR APPLICATION OF THE SOFTWARE AND DOCUMENTATION.
 * IN NO EVENT SHALL MICROCHIP OR ITS LICENSORS BE LIABLE OR OBLIGATED
 * UNDER CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH
 * OF WARRANTY, OR OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT
 * DAMAGES OR EXPENSES INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL,
 * SPECIAL, INDIRECT, PUNITIVE OR CONSEQUENTIAL DAMAGES, LOST PROFITS
 * OR LOST DATA, COST OF PROCUREMENT OF SUBSTITUTE GOODS, TECHNOLOGY,
 * SERVICES, OR ANY CLAIMS BY THIRD PARTIES (INCLUDING BUT NOT LIMITED
 * TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.
 *
 ********************************************************************/

#include "RTLL.h"
#include <string.h>

/********************************************************************
* Function: 	dlopen()
*
* Precondition: 
*
* Input: 		*libString: Pointer to Library/Module Name.
				*libLoadPoint: Address of the library/module header.
                *libOpenParam: Pointer to a parameter to be passed to
                               library init routine
*
* Output:		Handle to library header.
*
* Side Effects:	None.
*
* Overview:      Does following things.
				* Checks if library is loaded.
				* Calls the library start up code, if library is found.
*
*			
* Note:		    This function should be called before any other 
*               RTLL function is called.
********************************************************************/
void* dlopen(char *libString, void* libLoadPoint, void * libOpenParam)
{
	
	const T_MODULE_DYN_HDR* lib = (const T_MODULE_DYN_HDR*)libLoadPoint;
	
	if(!strcmp(lib->module_name, libString))
	{
		// Got the library we are looking for
		lib->libInit((void*)libOpenParam);
		return((void*)lib);	
	}	
	// Failed to get the library.
	return 0;	
}	


/********************************************************************
* Function: 	dlsym()
*
* Precondition: 
*
* Input: 		*libHandle: Handle to the library header.
				*funcName: Proc/func name string.
*
* Output:		Proc Address.
*
* Side Effects:	None.
*
* Overview:      Does following things.
				* Checks if library has the function/procedure.
				*Returns the function/procedure handle if the
				string is found in the export proc table of the
				library.
*
*			
* Note:		 	This handle must be called by the application 
				to initialize the library.
********************************************************************/
void*  dlsym(void* libHandle, char* funcName)
{
	const T_MODULE_DYN_HDR* libPtr = (const T_MODULE_DYN_HDR*)libHandle;
	int i;
	
	
	for(i = 0; i < libPtr->nProcs; i++)
	{
		if(!strcmp(libPtr->procTbl[i].procString, funcName))
		{
			return ((void *)libPtr->procTbl[i].proc);	
		}	
		
	}		
	return 0;	
}	



